from __future__ import annotations

from typing import Dict

from pyrepositoryminer.metrics.tree import TreeMetric, TreeMetricOutput
from pyrepositoryminer.visitableobject import VisitableBlob


class Filecount(TreeMetric):
    def __init__(self, cache: Dict[str, bool]) -> None:
        super().__init__(cache)
        self.n: int = 0

    def visitBlob(self, blob: VisitableBlob) -> Filecount:
        self.n += 1
        return self

    @property
    def result(self) -> TreeMetricOutput:
        return TreeMetricOutput(value=self.n)



from __future__ import annotations

from typing import Dict, Iterable, List

from radon.raw import analyze

from pyrepositoryminer.metrics.blob import BlobMetric, BlobMetricOutput
from pyrepositoryminer.visitableobject import VisitableBlob


class Raw(BlobMetric):
    def __init__(self, cache: Dict[str, bool]) -> None:
        super().__init__(cache)
        self.metrics: List[BlobMetricOutput] = []

    def visitBlob(self, blob: VisitableBlob) -> Raw:
        if not blob.obj.name.endswith(".py"):
            return self
        id = str(blob.obj.id)
        if self.cached_oids.setdefault(id, False):
            self.metrics.append(
                BlobMetricOutput(value="-", blob_id=id, blob_name=self.pathname)
            )
            return self
        self.cached_oids[id] = True
        self.metrics.append(
            BlobMetricOutput(
                value=analyze(blob.obj.data.decode())._asdict(),
                blob_id=id,
                blob_name=self.pathname,
            )
        )
        return self

    @property
    def result(self) -> Iterable[BlobMetricOutput]:
        return self.metrics



from __future__ import annotations

from typing import Dict, Iterable, List

from radon.metrics import h_visit

from pyrepositoryminer.metrics.unit import UnitMetric, UnitMetricOutput
from pyrepositoryminer.visitableobject import VisitableBlob


class Halstead(UnitMetric):
    def __init__(self, cache: Dict[str, bool]) -> None:
        super().__init__(cache)
        self.metrics: List[UnitMetricOutput] = []

    def visitBlob(self, blob: VisitableBlob) -> Halstead:
        if blob.obj.name.endswith(".py"):
            self.metrics.extend(
                UnitMetricOutput(
                    unit_id=str(function_name),
                    value=report._asdict(),
                    blob_id=blob.obj.id,
                    blob_name=self.pathname,
                )
                for function_name, report in h_visit(blob.obj.data).functions
            )

        return self

    @property
    def result(self) -> Iterable[UnitMetricOutput]:
        return self.metrics



from enum import Enum
from itertools import chain
from json import dumps
from multiprocessing import Manager, Pool
from pathlib import Path
from sys import stdin
from typing import Dict, Iterable, List, Optional, Set, Tuple

from pygit2 import (
    GIT_SORT_NONE,
    GIT_SORT_REVERSE,
    GIT_SORT_TIME,
    GIT_SORT_TOPOLOGICAL,
    Commit,
    Repository,
    Walker,
    clone_repository,
)
from typer import Argument, FileText, Option, Typer, echo

from pyrepositoryminer.analyze import analyze as analyze_worker
from pyrepositoryminer.analyze import initialize as initialize_worker
from pyrepositoryminer.metrics import BlobMetrics, TreeMetrics, UnitMetrics

app = Typer(help="Efficient Repository Mining in Python.")

AvailableMetrics = Enum(  # type: ignore
    "AvailableMetrics",
    sorted((key, key) for key in {*TreeMetrics, *BlobMetrics, *UnitMetrics}),
)


class Sort(str, Enum):
    topological = "topological"
    time = "time"


def validate_metrics(
    metrics: Optional[List[AvailableMetrics]],
) -> Tuple[Set[str], Set[str], Set[str]]:
    if metrics is None:
        return (set(), set(), set())
    distinct = {metric.value for metric in metrics}
    return (
        distinct & TreeMetrics.keys(),
        distinct & BlobMetrics.keys(),
        distinct & UnitMetrics.keys(),
    )


def walk_commits(
    repo: Repository, branch_name: str, simplify_first_parent: bool
) -> Iterable[Commit]:
    branch = repo.branches[branch_name]
    walker: Walker = repo.walk(
        branch.peel().id, GIT_SORT_TOPOLOGICAL | GIT_SORT_REVERSE
    )
    if simplify_first_parent:
        walker.simplify_first_parent()
    yield from walker


@app.command()
def commits(
    repository: Path = Argument(..., help="The path to the repository."),
    branches: Optional[FileText] = Option(
        None, help="The branches to pull the commits from."
    ),
    simplify_first_parent: bool = True,
    drop_duplicates: bool = False,
    sort: Optional[Sort] = Option(None, case_sensitive=False),
    sort_reverse: bool = False,
) -> None:
    """Get the commit ids of a repository."""
    repo = Repository(repository)
    sorting = GIT_SORT_NONE
    if sort == "topological":
        sorting = GIT_SORT_TOPOLOGICAL
    elif sort == "time":
        sorting = GIT_SORT_TIME
    if sort_reverse:
        sorting |= GIT_SORT_REVERSE
    walkers = []
    for branch_name in (
        branch_name.strip() for branch_name in (stdin if branches is None else branches)
    ):
        branch = repo.branches[branch_name]
        walker: Walker = repo.walk(branch.peel().id)
        if simplify_first_parent:
            walker.simplify_first_parent()
        walkers.append(walker)
    all_commits: Iterable[Commit] = chain(*walkers)
    if drop_duplicates:
        all_commits = set(all_commits)
    for commit in all_commits:
        echo(str(commit.id))


@app.command()
def analyze(
    repository: Path,
    metrics: Optional[List[AvailableMetrics]] = Argument(None, case_sensitive=False),
    commits: Optional[FileText] = None,
    workers: int = 1,
) -> None:
    """Analyze commits of a repository."""
    workers = max(workers, 1)
    tree_m, blob_m, unit_m = validate_metrics(metrics)

    manager = Manager()
    cache: Dict[str, bool] = manager.dict()
    with Pool(
        max(workers, 1), initialize_worker, (repository, tree_m, blob_m, unit_m, cache)
    ) as pool:
        for result in pool.imap(
            analyze_worker,
            (
                commit_id.strip()
                for commit_id in (stdin if commits is None else commits)
            ),
        ):
            if result is not None:
                echo(dumps(result, separators=(",", ":"), indent=None))


@app.command()
def clone(
    url: str,
    path: Path,
) -> None:
    "Clone a repository to a path."
    clone_repository(url, path, bare=True)


@app.command()
def branch(path: Path, local: bool = True, remote: bool = True) -> None:
    """Get the branches of a repository."""
    repo = Repository(path)
    branches: Iterable[str] = tuple()
    if local and remote:
        branches = repo.branches
    elif local:
        branches = repo.branches.local
    elif remote:
        branches = repo.branches.remote
    for branch_name in branches:
        echo(branch_name)



from typing import Dict, Type

from pyrepositoryminer.metrics.blob import BlobMetric
from pyrepositoryminer.metrics.blob.halstead import Halstead as HalsteadTotal
from pyrepositoryminer.metrics.blob.maintainability import Maintainability
from pyrepositoryminer.metrics.blob.nesting import Nesting
from pyrepositoryminer.metrics.blob.raw import Raw
from pyrepositoryminer.metrics.tree import TreeMetric
from pyrepositoryminer.metrics.tree.filecount import Filecount
from pyrepositoryminer.metrics.tree.loc import LOC
from pyrepositoryminer.metrics.unit import UnitMetric
from pyrepositoryminer.metrics.unit.complexity import Complexity
from pyrepositoryminer.metrics.unit.halstead import Halstead
from pyrepositoryminer.metrics.unit.linelength import Linelength

TreeMetrics: Dict[str, Type[TreeMetric]] = {"filecount": Filecount, "loc": LOC}
BlobMetrics: Dict[str, Type[BlobMetric]] = {
    "nesting": Nesting,
    "raw": Raw,
    "halstead_total": HalsteadTotal,
    "maintainability": Maintainability,
}
UnitMetrics: Dict[str, Type[UnitMetric]] = {
    "complexity": Complexity,
    "linelength": Linelength,
    "halstead": Halstead,
}



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Iterable

from pyrepositoryminer.metrics.blob import BlobMetricOutput
from pyrepositoryminer.metrics.visitor import TreeVisitor


@dataclass(frozen=True)
class UnitMetricOutput(BlobMetricOutput):
    unit_id: str


class UnitMetric(TreeVisitor, ABC):
    @property
    @abstractmethod
    def result(self) -> Iterable[UnitMetricOutput]:
        pass



from __future__ import annotations

from ast import (
    AST,
    Break,
    Continue,
    ExceptHandler,
    For,
    If,
    NodeVisitor,
    Try,
    While,
    With,
    parse,
    withitem,
)
from typing import Dict, Iterable, List

from pyrepositoryminer.metrics.blob import BlobMetric, BlobMetricOutput
from pyrepositoryminer.visitableobject import VisitableBlob


class NestingASTVisitor(NodeVisitor):
    def __init__(self) -> None:
        super().__init__()
        self.n: int = 0
        self.max_n: int = self.n

    def visit(self, node: AST) -> NestingASTVisitor:
        self.max_n = max(self.max_n, self.n)
        super().visit(node)
        return self

    def nesting_visit(self, node: AST) -> None:
        self.n += 1
        super().generic_visit(node)
        self.n -= 1

    def visit_If(self, node: If) -> None:
        return self.nesting_visit(node)

    def visit_For(self, node: For) -> None:
        return self.nesting_visit(node)

    def visit_While(self, node: While) -> None:
        return self.nesting_visit(node)

    def visit_Break(self, node: Break) -> None:
        return self.nesting_visit(node)

    def visit_Continue(self, node: Continue) -> None:
        return self.nesting_visit(node)

    def visit_Try(self, node: Try) -> None:
        return self.nesting_visit(node)

    def visit_ExceptHandler(self, node: ExceptHandler) -> None:
        return self.nesting_visit(node)

    def visit_With(self, node: With) -> None:
        return self.nesting_visit(node)

    def visit_withitem(self, node: withitem) -> None:
        return self.nesting_visit(node)

    @property
    def result(self) -> int:
        return self.max_n


class Nesting(BlobMetric):
    def __init__(self, cache: Dict[str, bool]) -> None:
        super().__init__(cache)
        self.metrics: List[BlobMetricOutput] = []

    def visitBlob(self, blob: VisitableBlob) -> Nesting:
        if blob.obj.name.endswith(".py"):
            m = parse(blob.obj.data)
            self.metrics.append(
                BlobMetricOutput(
                    value=NestingASTVisitor().visit(m).result,
                    blob_id=blob.obj.id,
                    blob_name=self.pathname,
                )
            )
        return self

    @property
    def result(self) -> Iterable[BlobMetricOutput]:
        return self.metrics



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from pygit2 import Blob, Object, Tree

if TYPE_CHECKING:
    from pyrepositoryminer.metrics.visitor import TreeVisitor


class VisitableObject(ABC):
    @abstractmethod
    def __init__(self, obj: Object) -> None:
        pass

    @abstractmethod
    def accept(self, treevisitor: "TreeVisitor") -> None:
        pass


class VisitableTree(VisitableObject):
    def __init__(self, obj: Tree) -> None:
        super().__init__(obj)
        self.obj: Tree = obj

    def accept(self, treevisitor: "TreeVisitor") -> None:
        treevisitor.visitTree(self)


class VisitableBlob(VisitableObject):
    def __init__(self, obj: Blob) -> None:
        super().__init__(obj)
        self.obj: Blob = obj

    def accept(self, treevisitor: "TreeVisitor") -> None:
        treevisitor.visitBlob(self)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, List

from pygit2 import Blob, Tree

from pyrepositoryminer.visitableobject import VisitableBlob, VisitableTree


class TreeVisitor(ABC):
    def __init__(self, cache: Dict[str, bool]) -> None:
        self.cached_oids = cache
        self.path: List[str] = []

    @property
    def pathname(self) -> str:
        return "".join(self.path)

    @abstractmethod
    def visitBlob(self, blob: VisitableBlob) -> TreeVisitor:
        pass

    def visitTree(self, tree: VisitableTree) -> TreeVisitor:
        for obj in tree.obj:
            id = str(obj.id)
            if isinstance(obj, Tree):
                self.path.append(f"{obj.name}/")
                self.cached_oids.setdefault(id, False)
                VisitableTree(obj).accept(self)
                self.cached_oids[id] = True
            elif isinstance(obj, Blob):
                self.path.append(str(obj.name))
                VisitableBlob(obj).accept(self)
            self.path.pop(-1)
        return self

    @property
    @abstractmethod
    def result(self) -> Any:
        pass



"""Analyze a commit w.r.t tree-, blob- or unit-metrics.

Global variables are accessed in the context of a worker.
"""
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Tuple, TypedDict

from pygit2 import Commit, Repository, Signature, Tree

from pyrepositoryminer.metrics import BlobMetrics, TreeMetrics, UnitMetrics
from pyrepositoryminer.visitableobject import VisitableTree

repo: Repository

tm: Tuple[str, ...]
bm: Tuple[str, ...]
um: Tuple[str, ...]
cached_oids: Dict[str, bool]


class Metric(TypedDict):
    name: str
    value: Any


class ObjectOutput(TypedDict):
    id: str


class UnitOutput(ObjectOutput):
    metrics: List[Metric]


class BlobOutput(ObjectOutput):
    name: str
    metrics: List[Metric]
    units: List[UnitOutput]


class SignatureOutput(TypedDict):
    email: str
    name: str
    time_offset: int
    time: int


class CommitBase(ObjectOutput):
    author: SignatureOutput
    commit_time: int
    commit_time_offset: int
    committer: SignatureOutput
    message: str
    parent_ids: List[str]


class CommitOutput(CommitBase, total=False):
    metrics: List[Metric]
    blobs: List[BlobOutput]


BlobEntry = TypedDict(
    "BlobEntry",
    {"name": str, "metrics": List[Metric], "units": Dict[str, List[Metric]]},
)


def parse_commit(
    commit: Commit,
    metrics: Iterable[Metric],
    blobs: Iterable[BlobOutput],
) -> CommitOutput:
    def parse_signature(signature: Signature) -> SignatureOutput:
        return SignatureOutput(
            email=str(signature.email),
            name=str(signature.name),
            time_offset=int(signature.offset),
            time=int(signature.time),
        )

    return CommitOutput(
        id=str(commit.id),
        author=parse_signature(commit.author),
        commit_time=int(commit.commit_time),
        commit_time_offset=int(commit.commit_time_offset),
        committer=parse_signature(commit.committer),
        message=str(commit.message),
        parent_ids=[str(id) for id in commit.parent_ids],
        metrics=list(metrics),
        blobs=list(blobs),
    )


def validate_commit(repository: Repository, commit_id: str) -> bool:
    try:
        obj = repository.get(commit_id)
    except ValueError:
        return False
    else:
        return obj is not None and isinstance(obj, Commit)


def initialize(
    repository: Path,
    tree_m: Iterable[str],
    blob_m: Iterable[str],
    unit_m: Iterable[str],
    cache: Dict[str, bool],
) -> None:
    global repo, tm, bm, um, cached_oids
    repo = Repository(repository)
    tm = tuple(sorted(tree_m))
    bm = tuple(sorted(blob_m))
    um = tuple(sorted(unit_m))
    cached_oids = cache


def analyze_unit(tree: Tree) -> Iterable[Tuple[str, str, str, Metric]]:
    global um, cached_oids
    for m in um:
        for res in UnitMetrics[m](cached_oids).visitTree(VisitableTree(tree)).result:
            yield str(res.blob_id), str(res.blob_name), str(res.unit_id), Metric(
                name=m, value=res.value
            )


def analyze_blob(tree: Tree) -> Iterable[Tuple[str, str, Metric]]:
    global bm, cached_oids
    for m in bm:
        for res in BlobMetrics[m](cached_oids).visitTree(VisitableTree(tree)).result:
            yield str(res.blob_id), str(res.blob_name), Metric(name=m, value=res.value)


def analyze_tree(tree: Tree) -> Iterable[Metric]:
    global tm, cached_oids
    for m in tm:
        yield Metric(
            name=m,
            value=TreeMetrics[m](cached_oids)
            .visitTree(VisitableTree(tree))
            .result.value,
        )


def analyze(commit_id: str) -> Optional[CommitOutput]:
    global repo
    try:
        commit = repo.get(commit_id)
    except ValueError:
        return None
    else:
        if commit is None or not isinstance(commit, Commit):
            return None

    d: Dict[str, BlobEntry] = {}
    for blob_id, blob_name, unit_id, metric in analyze_unit(commit.tree):
        d.setdefault(blob_id, {"name": blob_name, "metrics": [], "units": {}})[
            "units"
        ].setdefault(unit_id, []).append(metric)
    for blob_id, blob_name, metric in analyze_blob(commit.tree):
        d.setdefault(blob_id, {"name": blob_name, "metrics": [], "units": {}})[
            "metrics"
        ].append(metric)
    return parse_commit(
        commit,
        metrics=analyze_tree(commit.tree),
        blobs=[
            BlobOutput(
                id=str(blob_id),
                name=blob["name"],
                metrics=blob["metrics"],
                units=[
                    UnitOutput(id=unit_id, metrics=unit)
                    for unit_id, unit in blob["units"].items()
                ],
            )
            for blob_id, blob in d.items()
        ],
    )



from __future__ import annotations

from typing import Dict, Iterable, List

from pyrepositoryminer.metrics.unit import UnitMetric, UnitMetricOutput
from pyrepositoryminer.visitableobject import VisitableBlob


class Linelength(UnitMetric):
    def __init__(self, cache: Dict[str, bool]) -> None:
        super().__init__(cache)
        self.metrics: List[UnitMetricOutput] = []

    def visitBlob(self, blob: VisitableBlob) -> Linelength:
        if blob.obj.is_binary:
            return self
        for i, line in enumerate(blob.obj.data.split(b"\n")):
            self.metrics.append(
                UnitMetricOutput(
                    len(line),
                    blob_id=blob.obj.id,
                    unit_id=f"L{i+1}",
                    blob_name=self.pathname,
                )
            )
        return self

    @property
    def result(self) -> Iterable[UnitMetricOutput]:
        return self.metrics



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Iterable

from pyrepositoryminer.metrics.tree import TreeMetricOutput
from pyrepositoryminer.metrics.visitor import TreeVisitor


@dataclass(frozen=True)
class BlobMetricOutput(TreeMetricOutput):
    blob_id: str
    blob_name: str


class BlobMetric(TreeVisitor, ABC):
    @property
    @abstractmethod
    def result(self) -> Iterable[BlobMetricOutput]:
        pass



from __future__ import annotations

from typing import Dict, Iterable, List

from radon.metrics import mi_visit

from pyrepositoryminer.metrics.blob import BlobMetric, BlobMetricOutput
from pyrepositoryminer.visitableobject import VisitableBlob


class Maintainability(BlobMetric):
    def __init__(self, cache: Dict[str, bool]) -> None:
        super().__init__(cache)
        self.metrics: List[BlobMetricOutput] = []

    def visitBlob(self, blob: VisitableBlob) -> Maintainability:
        if blob.obj.name.endswith(".py"):
            self.metrics.append(
                BlobMetricOutput(
                    value=mi_visit(blob.obj.data.decode(), multi=True),
                    blob_id=blob.obj.id,
                    blob_name=self.pathname,
                )
            )

        return self

    @property
    def result(self) -> Iterable[BlobMetricOutput]:
        return self.metrics



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyrepositoryminer',
 'pyrepositoryminer.metrics',
 'pyrepositoryminer.metrics.blob',
 'pyrepositoryminer.metrics.tree',
 'pyrepositoryminer.metrics.unit']

package_data = \
{'': ['*']}

install_requires = \
['pygit2>=1.5.0,<2.0.0', 'radon>=4.5.0,<5.0.0', 'typer[all]>=0.3.2,<0.4.0']

entry_points = \
{'console_scripts': ['pyrepositoryminer = pyrepositoryminer.main:app']}

setup_kwargs = {
    'name': 'pyrepositoryminer',
    'version': '0.3.0',
    'description': 'Efficient Repository Mining in Python',
    'long_description': '# pyrepositoryminer\n\n[![CI workflow](https://github.com/fabianhe/pyrepositoryminer/actions/workflows/test.yaml/badge.svg)](https://github.com/fabianhe/pyrepositoryminer/actions/workflows/test.yaml)\n[![PyPI](https://img.shields.io/pypi/v/pyrepositoryminer?color=%23000)](https://pypi.org/project/pyrepositoryminer/)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\nEfficient Repository Mining in Python\n\n## Development Setup\n\n1. Install [poetry](https://github.com/python-poetry/poetry)\n   ```bash\n   curl -sSL https://raw.githubusercontent.com/python-poetry/poetry/master/install-poetry.py | python -\n   ```\n2. Install the dependencies\n   ```bash\n   poetry install\n   ```\n3. Install the [pre-commit](https://github.com/pre-commit/pre-commit) hooks\n   ```bash\n   pre-commit install\n   ```\n',
    'author': 'Fabian Heseding',
    'author_email': '39628987+fabianhe@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/fabianhe/pyrepositoryminer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any

from pyrepositoryminer.metrics.visitor import TreeVisitor


@dataclass(frozen=True)
class TreeMetricOutput:
    value: Any


class TreeMetric(TreeVisitor, ABC):
    @property
    @abstractmethod
    def result(self) -> TreeMetricOutput:
        pass



from __future__ import annotations

from typing import Dict, Iterable, List

from radon.complexity import cc_visit

from pyrepositoryminer.metrics.unit import UnitMetric, UnitMetricOutput
from pyrepositoryminer.visitableobject import VisitableBlob


class Complexity(UnitMetric):
    def __init__(self, cache: Dict[str, bool]) -> None:
        super().__init__(cache)
        self.metrics: List[UnitMetricOutput] = []

    def visitBlob(self, blob: VisitableBlob) -> Complexity:
        if blob.obj.name.endswith(".py"):
            if complexities := cc_visit(blob.obj.data):
                self.metrics.extend(
                    UnitMetricOutput(
                        unit_id=str(u.fullname),
                        value=int(u.complexity),
                        blob_id=blob.obj.id,
                        blob_name=self.pathname,
                    )
                    for u in complexities
                )

        return self

    @property
    def result(self) -> Iterable[UnitMetricOutput]:
        return self.metrics
